@@ -102,6 +102,12 @@ typedef enum {
     ISPX,
 } amdgpu_pro_px_action;
 
+typedef enum {
+    ON,
+    OFF,
+    ONDEMAND
+} prime_mode_settings;
+
 static char *log_file = NULL;
 static FILE *log_handle = NULL;
 static char *last_boot_file = NULL;
@@ -116,6 +122,7 @@ static char *amdgpu_pro_px_file = NULL;
 static char *modprobe_d_path = NULL;
 static char *xorg_conf_d_path = NULL;
 static prime_intel_drv prime_intel_driver = SNA;
+static prime_mode_settings prime_mode = OFF;
 
 static struct pci_slot_match match = {
     PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY, 0
@@ -518,44 +525,34 @@ static bool copy_file(const char *src_path, const char *dst_path)
 }
 
 
-/* Open a file and check if it contains "on"
- * or "off".
- *
- * Return false if the file doesn't exist or is empty.
- */
-static bool check_on_off(const char *path) {
-    bool status = false;
+/* Get prime action, which can be "on", "off", or "on-demand" */
+static void get_prime_action() {
     char line[100];
     _cleanup_fclose_ FILE *file = NULL;
 
-    file = fopen(path, "r");
+    file = fopen(prime_settings, "r");
 
     if (!file) {
-        fprintf(log_handle, "Error: can't open %s\n", path);
-        return false;
+        fprintf(log_handle, "Error: can't open %s\n", prime_settings);
+        prime_mode = OFF;
     }
 
     while (fgets(line, sizeof(line), file)) {
-        if (istrstr(line, "on") != NULL) {
-            status = true;
+        if (istrstr(line, "on-demand") != NULL) {
+            prime_mode = ONDEMAND;
+            break;
+        }
+        else if (istrstr(line, "on") != NULL) {
+            prime_mode = ON;
+            break;
+        }
+        else {
+            prime_mode = OFF;
             break;
         }
     }
-
-    return status;
 }
 
-
-/* Get the settings for PRIME.
- *
- * This tells us whether the discrete card should be
- * on or off.
- */
-static bool prime_is_action_on() {
-    return (check_on_off(prime_settings));
-}
-
-
 static void get_boot_vga(struct device **devices,
                         int cards_number,
                         unsigned int *vendor_id,
@@ -997,7 +994,7 @@ static bool is_connector_connected(const char *connector) {
 
 /* Count the number of outputs connected to the card */
 int count_connected_outputs(const char *device_name) {
-    char name[50];
+    char name[PATH_MAX];
     struct dirent *dp;
     DIR *dfd;
     int connected_outputs = 0;
@@ -1016,7 +1013,8 @@ int count_connected_outputs(const char *device_name) {
                     drm_dir, dp->d_name);
         else {
             /* Open the file for the connector */
-            sprintf(name, "%s/%s/status", drm_dir, dp->d_name);
+            snprintf(name, sizeof(name), "%s/%s/status", drm_dir, dp->d_name);
+            name[sizeof(name) - 1] = 0;
             if (is_connector_connected(name)) {
                 fprintf(log_handle, "output %d:\n", connected_outputs);
                 fprintf(log_handle, "\t%s\n", dp->d_name);
@@ -1036,7 +1034,7 @@ int count_connected_outputs(const char *device_name) {
 static int has_driver_connected_outputs(const char *driver) {
     DIR *dir;
     struct dirent* dir_entry;
-    char path[20];
+    char path[PATH_MAX];
     int fd = 1;
     drmVersionPtr version;
     int connected_outputs = 0;
@@ -1055,6 +1053,7 @@ static int has_driver_connected_outputs(const char *driver) {
             continue;
 
         snprintf(path, sizeof(path), "%s/%s", dri_dir, dir_entry->d_name);
+        path[sizeof(path) - 1] = 0;
         fd = open(path, O_RDWR);
         if (fd) {
             if ((version = drmGetVersion(fd))) {
@@ -1376,6 +1375,33 @@ static bool create_prime_outputclass(void) {
     return false;
 }
 
+static bool create_offload_serverlayout(void) {
+    _cleanup_fclose_ FILE *file = NULL;
+    char xorg_d_custom[PATH_MAX];
+
+    snprintf(xorg_d_custom, sizeof(xorg_d_custom), "%s/11-nvidia-offload.conf",
+             xorg_conf_d_path);
+
+    fprintf(log_handle, "Creating %s\n", xorg_d_custom);
+    file = fopen(xorg_d_custom, "w");
+    if (!file) {
+        fprintf(log_handle, "Error while creating %s\n", xorg_d_custom);
+    }
+    else {
+        fprintf(file,
+                "# DO NOT EDIT. AUTOMATICALLY GENERATED BY gpu-manager\n\n"
+                "Section \"ServerLayout\"\n"
+                "    Identifier \"layout\"\n"
+                "    Option \"AllowNVIDIAGPUScreens\"\n"
+                "EndSection\n\n");
+
+        fflush(file);
+        return true;
+    }
+
+    return false;
+}
+
 static void remove_prime_outputclass(void) {
     char xorg_d_custom[PATH_MAX];
     snprintf(xorg_d_custom, sizeof(xorg_d_custom), "%s/11-nvidia-prime.conf",
@@ -1384,6 +1410,13 @@ static void remove_prime_outputclass(void) {
     unlink(xorg_d_custom);
 }
 
+static void remove_offload_serverlayout(void) {
+    char xorg_d_custom[PATH_MAX];
+    snprintf(xorg_d_custom, sizeof(xorg_d_custom), "%s/11-nvidia-offload.conf",
+            xorg_conf_d_path);
+    fprintf(log_handle, "Removing %s\n", xorg_d_custom);
+    unlink(xorg_d_custom);
+}
 
 static bool manage_power_management(const struct device *device, bool enabled) {
     _cleanup_fclose_ FILE *file = NULL;
@@ -1410,11 +1443,11 @@ static bool manage_power_management(const struct device *device, bool enabled) {
     }
 }
 
-static bool enable_power_management(const struct device *device) {
+static void enable_power_management(const struct device *device) {
     manage_power_management(device, true);
 }
 
-static bool disable_power_management(const struct device *device) {
+static void disable_power_management(const struct device *device) {
     manage_power_management(device, false);
 }
 
@@ -1617,18 +1650,33 @@ static bool enable_prime(const char *prime_settings,
         }
     }
 
-    if (prime_is_action_on()) {
+    get_prime_action();
+    if (prime_mode == ON) {
         /* Create an OutputClass just for PRIME, to override
          * the default NVIDIA settings
          */
         create_prime_outputclass();
+        /* Remove the ServerLayout */
+        remove_offload_serverlayout();
         disable_power_management(device);
         if (!is_module_loaded("nvidia"))
             load_module("nvidia");
     }
-    else {
+    else if (prime_mode == ONDEMAND) {
+        /* Create the ServerLayout required to enabling offload
+         * for NVIDIA.
+         */
+        create_offload_serverlayout();
         /* Remove the OutputClass */
         remove_prime_outputclass();
+        enable_power_management(device);
+        if (!is_module_loaded("nvidia"))
+            load_module("nvidia");
+    }
+    else {
+        /* Remove the OutputClass and ServerLayout */
+        remove_prime_outputclass();
+        remove_offload_serverlayout();
 
 unload_again:
         /* Unload the NVIDIA modules and enable pci power management */
